from copy import deepcopy
from typing import List, Dict, Union, Optional, Literal, Any, overload

from ural import is_url

from twitwi.exceptions import BlueskyPayloadError
from twitwi.utils import (
    get_collection_time,
    get_dates,
    safe_normalize_url,
    custom_get_normalized_hostname,
)
from twitwi.bluesky.utils import (
    validate_post_payload,
    valid_embed_type,
    format_profile_url,
    format_post_url,
    parse_post_url,
    parse_post_uri,
    format_starterpack_url,
    format_media_url,
)
from twitwi.bluesky.types import BlueskyProfile, BlueskyPartialProfile, BlueskyPost


def normalize_profile(data: Dict, locale: Optional[Any] = None) -> BlueskyProfile:
    associated = data["associated"]

    pinned_post_uri = None
    pinned_post_data = data.get("pinnedPost")

    if pinned_post_data is not None:
        pinned_post_uri = pinned_post_data["uri"]

    timestamp_utc, created_at = get_dates(
        data["createdAt"], locale=locale, source="bluesky"
    )

    return {
        "did": data["did"],
        "url": format_profile_url(data["handle"]),
        "handle": data["handle"],
        "display_name": data.get("displayName"),
        "created_at": created_at,
        "timestamp_utc": timestamp_utc,
        "description": data.get("description"),
        "avatar": data.get("avatar"),
        "posts": data["postsCount"],
        "followers": data["followersCount"],
        "follows": data["followsCount"],
        "lists": associated["lists"],
        "feedgens": associated["feedgens"],
        "starter_packs": associated["starterPacks"],
        "banner": data.get("banner"),
        "pinned_post_uri": pinned_post_uri,
        "collection_time": get_collection_time(),
    }


def normalize_partial_profile(
    data: Dict, locale: Optional[Any] = None
) -> BlueskyPartialProfile:
    associated = data["associated"]

    timestamp_utc, created_at = get_dates(
        data["createdAt"], locale=locale, source="bluesky"
    )

    return {
        "did": data["did"],
        "url": format_profile_url(data["handle"]),
        "handle": data["handle"],
        "display_name": data.get("displayName"),
        "created_at": created_at,
        "timestamp_utc": timestamp_utc,
        "description": data.get("description"),
        "avatar": data.get("avatar"),
        "lists": associated.get("lists"),
        "feedgens": associated.get("feedgens"),
        "starter_packs": associated.get("starterPacks"),
        "collection_time": get_collection_time(),
    }


def prepare_native_gif_as_media(gif_data, user_did, source):
    if "thumb" in gif_data:
        media_cid = gif_data["thumb"]["ref"]["$link"]
        _, thumbnail = format_media_url(user_did, media_cid, "image/jpeg", source)
    else:
        media_cid = ""
        thumbnail = ""

    return {
        "id": media_cid,
        "type": "video/gif",
        "alt": gif_data["title"],
        "url": gif_data["uri"],
        "thumb": thumbnail,
    }


def prepare_image_as_media(image_data):
    if "ref" not in image_data["image"] or "$link" not in image_data["image"]["ref"]:
        image_id = image_data["image"]["cid"]
    else:
        image_id = image_data["image"]["ref"]["$link"]
    return {
        "id": image_id,
        "type": image_data["image"]["mimeType"],
        "alt": image_data["alt"],
    }


def prepare_video_as_media(video_data):
    return {
        "id": video_data["ref"]["$link"],
        "type": video_data["mimeType"],
    }


def process_starterpack_card(embed_data, post):
    # Warning: mutates post

    card = embed_data.get("record", {})
    if "uri" in embed_data:
        creator_did, pack_did = parse_post_uri(embed_data["uri"])
        post["card_link"] = format_starterpack_url(
            embed_data.get("creator", {}).get("handle") or creator_did, pack_did
        )
    if card:
        post["card_title"] = card.get("name", "")
        post["card_description"] = card.get("description", "")
        post["card_thumbnail"] = card.get("thumb", "")
    return post


def process_card_data(embed_data, post):
    # Warning: mutates post

    post["card_link"] = embed_data["uri"]
    post["card_title"] = embed_data.get("title", "")
    post["card_description"] = embed_data.get("description", "")
    post["card_thumbnail"] = embed_data.get("thumb", "")
    return post


def prepare_quote_data(embed_quote, card_data, post, links):
    # Warning: mutates post and links

    quoted_data = None

    post["quoted_cid"] = embed_quote["cid"]
    post["quoted_uri"] = embed_quote["uri"]
    # Sometimes quoted post is not found, even if uri and cid are given
    # example: https://bsky.app/profile/takobiotech.masto.bike.ap.brid.gy/post/3lc6r7nzil6m2
    if card_data and card_data.get("notFound"):
        post["quoted_status"] = "notFound"
    else:
        post["quoted_user_did"], post["quoted_did"] = parse_post_uri(
            post["quoted_uri"], post["url"]
        )

        # First store ugly quoted url with user did in case full quote data is missing (recursion > 3 or detached quote)
        # Handling special posts types (only lists for now, for example: https://bsky.app/profile/lanana421.bsky.social/lists/3lxdgjtpqhf2z)
        if "/app.bsky.graph.list/" in post["quoted_uri"]:
            post_splitter = "/lists/"
        else:
            post_splitter = "/post/"
        post["quoted_url"] = format_post_url(
            post["quoted_user_did"], post["quoted_did"], post_splitter=post_splitter
        )

        if card_data:
            if card_data.get("detached"):
                post["quoted_status"] = "detached"

            else:
                quoted_data = deepcopy(card_data)

        # Grab user handle and cleanup links when no quote data but url in text
        if not quoted_data:
            for link in links:
                if link.startswith("https://bsky.app/profile/") and link.endswith(
                    post["quoted_did"]
                ):
                    # Take better quoted url with user_handle
                    post["quoted_url"] = link
                    break

            # Remove quoted link from post links
            if post["quoted_url"] in links:
                links.remove(post["quoted_url"])

            # Extract user handle from url
            if "did:plc:" not in post["quoted_url"]:
                post["quoted_user_handle"], _ = parse_post_url(
                    post["quoted_url"], post["url"]
                )

    return (post, quoted_data, links)


def merge_nested_posts(referenced_posts, nested, source):
    for new_post in nested:
        ordered_id = "%s_%s" % (new_post["did"], new_post["user_handle"])
        if ordered_id not in referenced_posts:
            referenced_posts[ordered_id] = new_post
        else:
            old_post = referenced_posts[ordered_id]
            for key in new_post.keys():
                if key not in old_post:
                    old_post[key] = new_post[key]
                elif old_post[key] != new_post[key]:
                    if key == "collected_via":
                        old_post[key] += new_post[key]
                    elif key == "match_query":
                        old_post[key] = old_post[key] or new_post[key]
                    elif key not in ["collection_time"]:
                        raise BlueskyPayloadError(
                            source,
                            "a nested post appearing twice in the same payload has some diverging metadata for key %s: %s / %s"
                            % (key, old_post[key], new_post[key]),
                        )
    return referenced_posts


@overload
def normalize_post(
    payload: Dict,
    locale: Optional[str] = ...,
    extract_referenced_posts: Literal[True] = ...,
    collection_source: Optional[str] = ...,
) -> List[BlueskyPost]: ...


@overload
def normalize_post(
    payload: Dict,
    locale: Optional[str] = ...,
    extract_referenced_posts: Literal[False] = ...,
    collection_source: Optional[str] = ...,
) -> BlueskyPost: ...


def normalize_post(
    payload: Dict,
    locale: Optional[Any] = None,
    extract_referenced_posts: bool = False,
    collection_source: Optional[str] = None,
) -> Union[BlueskyPost, List[BlueskyPost]]:
    """
    Function "normalizing" a post as returned by Bluesky's API in order to
    cleanup and optimize some fields.

    Args:
        payload (dict): post or feed payload json dict from Bluesky API.
        locale (pytz.timezone, optional): Timezone for date conversions.
        extract_referenced_posts (bool, optional): Whether to return, in
            addition to the original post, also the full list of posts
            found in the given payload (including the tree of quoted posts
            as well as the parent and root posts of the thread if the post
            comes as an answer to another one). Defaults
            to `False`.
        collection_source (str, optional): string explaining how the post
            was collected. Defaults to `None`.

    Returns:
        (dict or list): Either a single post dict or a list of post dicts if
            `extract_referenced_posts` was set to `True`.

    """

    if not isinstance(payload, dict):
        raise BlueskyPayloadError(
            "UNKNOWN", f"data provided to normalize_post is not a dictionary: {payload}"
        )

    valid, error = validate_post_payload(payload)
    if not valid:
        raise BlueskyPayloadError(
            payload.get("uri", payload.get("post", {}).get("uri", "UNKNOWN")),
            f"data provided to normalize_post is not a standard Bluesky post or feed payload:\n{error}",
        )

    if "post" in payload:
        data = payload["post"]
        reply_data = payload.get("reply")
        repost_data = payload.get("reason")
    else:
        data = payload
        reply_data = None
        repost_data = None

    if extract_referenced_posts:
        referenced_posts = {}

    if collection_source is None:
        collection_source = data.get("collection_source")

    post = {}

    # Store original text and prepare text for quotes & medias enriched version
    post["original_text"] = data["record"]["text"]
    text = post["original_text"].encode("utf-8")

    # Handle datetime fields
    post["collection_time"] = get_collection_time()
    post["timestamp_utc"], post["local_time"] = get_dates(
        data["record"]["createdAt"], locale=locale, source="bluesky"
    )
    post["indexed_at_utc"] = data["indexedAt"]

    # Handle post/user identifiers
    post["cid"] = data["cid"]
    post["uri"] = data["uri"]
    post["user_did"], post["did"] = parse_post_uri(data["uri"])
    post["user_handle"] = data["author"]["handle"]
    post["user_url"] = format_profile_url(post["user_handle"])
    post["url"] = format_post_url(post["user_handle"], post["did"])

    if post["user_did"] != data["author"]["did"]:
        raise BlueskyPayloadError(
            post["url"],
            "inconsistent user_did between Bluesky post's uri and post's author metadata: %s %s"
            % (data["uri"], data["author"]),
        )

    # Handle user metadata
    post["user_diplay_name"] = data["author"].get("displayName", "")
    post["user_avatar"] = data["author"].get("avatar", "")
    post["user_timestamp_utc"], post["user_created_at"] = get_dates(
        data["author"]["createdAt"], locale=locale, source="bluesky"
    )
    post["user_langs"] = data["record"].get("langs", [])

    if "bridgyOriginalUrl" in data["record"]:
        post["bridgy_original_url"] = data["record"]["bridgyOriginalUrl"]

    # Handle metrics
    post["repost_count"] = data["repostCount"]
    post["reply_count"] = data["replyCount"]
    post["like_count"] = data["likeCount"]
    post["quote_count"] = data["quoteCount"]
    # When a post cites another, the cited post doesn't have the bookmarkCount field
    post["bookmark_count"] = data.get("bookmarkCount")

    # Handle hashtags, mentions & links from facets
    post["mentioned_user_handles"] = []
    post["mentioned_user_dids"] = []
    hashtags = set()
    links = set()
    links_to_replace = []
    for facet in data["record"].get("facets", []):
        if len(facet["features"]) != 1:
            raise BlueskyPayloadError(
                post["url"],
                "unusual record facet content with more or less than a unique feature: %s"
                % facet,
            )

        feat = facet["features"][0]

        # Hashtags
        if feat["$type"].endswith("#tag") or feat["$type"].endswith("#hashtag"):
            hashtags.add(feat["tag"].strip().lower())

        # Mentions
        elif feat["$type"].endswith("#mention"):
            if feat["did"] not in post["mentioned_user_dids"]:
                post["mentioned_user_dids"].append(feat["did"])

                # Check & fix occasional errored mention positioning
                # example: https://bsky.app/profile/snjcgt.bsky.social/post/3lpmqkkkgp52u
                byteStart = facet["index"]["byteStart"]
                byteEnd = facet["index"]["byteEnd"]
                if text[byteStart : byteStart + 1] != b"@":
                    byteStart = text.find(b"@", byteStart)
                # in some cases, the errored positioning is before the position given
                # example: https://bsky.app/profile/springer.springernature.com/post/3lovyad4nt324
                if byteStart == -1 or byteStart > byteEnd:
                    # When decrementing byteStart, we also decrement byteEnd (see below)
                    # shifting the slice to extract the mention correctly
                    byteStart = facet["index"]["byteStart"] - 1
                    # to extend the size of the mention, which is somehow 1 char too short because of the '@'
                    byteEnd += 1

                handle = (
                    text[
                        byteStart + 1 : byteEnd
                        + byteStart
                        - facet["index"]["byteStart"]
                    ]
                    .strip()
                    .lower()
                    .decode("utf-8")
                )
                post["mentioned_user_handles"].append(handle)

        # Links
        elif feat["$type"].endswith("#link"):
            # Handle native polls
            if "https://poll.blue/" in feat["uri"]:
                if feat["uri"].endswith("/0"):
                    link = safe_normalize_url(feat["uri"])
                    if is_url(link):
                        links.add(link)
                    else:
                        continue
                    text += b" %s" % feat["uri"].encode("utf-8")
                continue

            link = safe_normalize_url(feat["uri"])
            if is_url(link):
                links.add(link)
            else:
                continue
            # Check & fix occasional errored link positioning
            # examples: https://bsky.app/profile/ecrime.ch/post/3lqotmopayr23
            #           https://bsky.app/profile/clustz.com/post/3lqfi7mnto52w
            byteStart = facet["index"]["byteStart"]
            byteEnd = facet["index"]["byteEnd"]

            if not text[byteStart:byteEnd].startswith(b"http"):
                new_byteStart = text.find(b"http", byteStart, byteEnd)

                # means that the link is shifted, like on this post:
                # https://bsky.app/profile/ecrime.ch/post/3lqotmopayr23
                if new_byteStart != -1:
                    byteStart = new_byteStart

                    # Find the index of the first space character after byteStart in case the link is a personalized one
                    # but still with the link in it (somehow existing in some posts, such as this one:
                    # https://bsky.app/profile/did:plc:rkphrshyfiqe4n2hz5vj56ig/post/3ltmljz5blca2)
                    # In this case, we don't want to touch the position of the link given in the payload
                    byteEnd = min(
                        byteStart
                        - facet["index"]["byteStart"]
                        + facet["index"]["byteEnd"],
                        len(post["original_text"].encode("utf-8")),
                    )
                    for i in range(byteStart, byteEnd):
                        if chr(text[i]).isspace():
                            byteStart = facet["index"]["byteStart"]
                    byteEnd = (
                        byteStart
                        - facet["index"]["byteStart"]
                        + facet["index"]["byteEnd"]
                    )

                # means that the link is a "personalized" one like on this post:
                # https://bsky.app/profile/newyork.activitypub.awakari.com.ap.brid.gy/post/3ln33tx7bpdu2
                else:
                    # we're looking for a link which could be valid if we add "https://" at the beginning,
                    # as in some cases the "http(s)://" part is missing in the post text
                    for starting in range(byteEnd - byteStart):
                        try:
                            if is_url(
                                "https://"
                                + text[
                                    byteStart + starting : byteEnd + starting
                                ].decode("utf-8")
                            ):
                                byteStart += starting
                                break
                        except UnicodeDecodeError:
                            pass
                    # If we did not find any valid link, we just keep the original position as it is
                    # meaning that we have a personalized link like in the example above

                    # Extend byteEnd to the right until we find a valid utf-8 ending,
                    # as in some cases the link is longer than the position given in the payload
                    # and it gets cut in the middle of a utf-8 char, leading to UnicodeDecodeError
                    # example: https://bsky.app/profile/radiogaspesie.bsky.social/post/3lmkzhvhtta22
                    while byteEnd <= len(post["original_text"].encode("utf-8")):
                        try:
                            text[byteStart:byteEnd].decode("utf-8")
                            break
                        except UnicodeDecodeError:
                            byteEnd += 1
                            continue

                    if byteEnd > len(post["original_text"].encode("utf-8")):
                        byteEnd = facet["index"]["byteEnd"]

                    byteEnd += byteStart - facet["index"]["byteStart"]

            # In some cases, the link is completely wrong in the post text,
            # like in this post: https://bsky.app/profile/sudetsoleil.bsky.social/post/3ljf3h74wee2m
            # So we chose to not replace anything in the text in this case
            try:
                text[byteStart:byteEnd].decode("utf-8")
                links_to_replace.append(
                    {
                        "uri": feat["uri"].encode("utf-8"),
                        "start": byteStart,
                        "end": byteEnd,
                    }
                )
            except UnicodeDecodeError:
                pass
                # raise UnicodeDecodeError(e.encoding, e.object, e.start, e.end, f"{e.reason} in post {post['url']}.\nText to decode: {text}\nSlice of text to decode: {text[e.start:e.end]}")

        elif feat["$type"].endswith("#bold"):
            pass
        elif feat["$type"].endswith("#option"):
            pass
        else:
            raise BlueskyPayloadError(
                post["url"], "unusual record facet feature $type: %s" % feat
            )
    post["hashtags"] = sorted(hashtags)

    # Rewrite full links within post's text
    for link in sorted(links_to_replace, key=lambda x: x["start"], reverse=True):
        if link["start"] < 0:
            text = text + b" " + link["uri"]
        else:
            text = text[: link["start"]] + link["uri"] + text[link["end"] :]

    # Handle thread info when applicable
    # Unfortunately posts' payload only provide at uris for these so we do not have the handles
    # We could sometimes resolve them from the mentionned and quote data but that would not handle most cases
    # Issue opened here to have user handles along: https://github.com/bluesky-social/atproto/issues/3722
    if "reply" in data["record"]:
        if "parent" in data["record"]["reply"]:
            post["to_post_cid"] = data["record"]["reply"]["parent"]["cid"]
            post["to_post_uri"] = data["record"]["reply"]["parent"]["uri"]
            post["to_user_did"], post["to_post_did"] = parse_post_uri(
                post["to_post_uri"], post["url"]
            )
            post["to_post_url"] = format_post_url(
                post["to_user_did"], post["to_post_did"]
            )
        if "root" in data["record"]["reply"]:
            post["to_root_post_cid"] = data["record"]["reply"]["root"]["cid"]
            post["to_root_post_uri"] = data["record"]["reply"]["root"]["uri"]
            post["to_root_user_did"], post["to_root_post_did"] = parse_post_uri(
                post["to_root_post_uri"], post["url"]
            )
            post["to_root_post_url"] = format_post_url(
                post["to_root_user_did"], post["to_root_post_did"]
            )

    # Handle quotes & medias
    media_ids = set()
    post["media_urls"] = []
    post["media_thumbnails"] = []
    post["media_types"] = []
    post["media_alt_texts"] = []
    if "embed" in data["record"]:
        embed = data["record"]["embed"]
        quoted_data = None
        media_data = []
        extra_links = []

        if not valid_embed_type(embed["$type"]):
            raise BlueskyPayloadError(
                post["url"], "unusual record embed $type: %s" % embed
            )

        # Links from cards
        if embed["$type"].endswith(".external"):
            link = embed["external"]["uri"]

            # Handle native gifs as medias
            if link.startswith("https://media.tenor.com/"):
                media_data.append(
                    prepare_native_gif_as_media(
                        embed["external"], post["user_did"], post["url"]
                    )
                )

            # Extra card links sometimes missing from facets & text due to manual action in post form
            else:
                extra_links.append(link)
                # Handle link card metadata
                if "embed" in data:
                    post = process_card_data(data["embed"]["external"], post)

        # Images
        if embed["$type"].endswith(".images"):
            media_data.extend([prepare_image_as_media(i) for i in embed["images"]])

        # Video
        if embed["$type"].endswith(".video"):
            media_data.append(prepare_video_as_media(embed["video"]))

        # Quote & Starter-packs
        if embed["$type"].endswith(".record"):
            if "app.bsky.graph.starterpack" in embed["record"]["uri"]:
                post = process_starterpack_card(
                    data.get("embed", {}).get("record", {}), post
                )
                if post.get("card_link"):
                    extra_links.append(post["card_link"])
            else:
                post, quoted_data, links = prepare_quote_data(
                    embed["record"], data.get("embed", {}).get("record"), post, links
                )

        # Quote with medias
        if embed["$type"].endswith(".recordWithMedia"):
            post, quoted_data, links = prepare_quote_data(
                embed["record"]["record"],
                data.get("embed", {}).get("record", {}).get("record"),
                post,
                links,
            )

            # Links from cards
            if embed["media"]["$type"].endswith(".external"):
                link = embed["media"]["external"]["uri"]

                # Handle native gifs as medias
                if link.startswith("https://media.tenor.com/"):
                    media_data.append(
                        prepare_native_gif_as_media(
                            embed["media"]["external"], post["user_did"], post["url"]
                        )
                    )

                # Extra card links sometimes missing from facets & text due to manual action in post form
                else:
                    extra_links = [link] + extra_links
                    # Handle link card metadata
                    if "embed" in data and "media" in data["embed"]["media"]:
                        post = process_card_data(
                            data["embed"]["media"]["external"], post
                        )

            # Images
            elif embed["media"]["$type"].endswith(".images"):
                media_data.extend(
                    [prepare_image_as_media(i) for i in embed["media"]["images"]]
                )

            # Video
            elif embed["media"]["$type"].endswith(".video"):
                media_data.append(prepare_video_as_media(embed["media"]["video"]))

            else:
                raise BlueskyPayloadError(
                    post["url"],
                    "unusual record embed media $type from a recordWithMedia: %s"
                    % embed,
                )

        # Process extra links
        for link in extra_links:
            norm_link = safe_normalize_url(link)
            if norm_link not in links:
                if is_url(norm_link):
                    links.add(norm_link)
                text += b" " + link.encode("utf-8")

        # Process medias
        for media in media_data:
            if media["id"] not in media_ids:
                media_ids.add(media["id"])
                media_type = media["type"]
                if "url" in media:
                    media_url = media["url"]
                    media_thumb = media["thumb"]
                else:
                    media_url, media_thumb = format_media_url(
                        post["user_did"], media["id"], media_type, post["url"]
                    )
                post["media_urls"].append(media_url)
                post["media_thumbnails"].append(media_thumb)
                post["media_types"].append(media_type)
                post["media_alt_texts"].append(media.get("alt", ""))

                # Rewrite post's text to include links to medias within
                text += b" " + (
                    media_thumb
                    if media_type.startswith("video")
                    and not media_type.endswith("/gif")
                    else media_url
                ).encode("utf-8")

        # Process quotes
        if quoted_data and "value" in quoted_data:
            # We're checking on the uri as the cid can be different in some cases,
            # and the uri seems to be unique for each post
            if quoted_data["uri"] != post["quoted_uri"]:
                raise BlueskyPayloadError(
                    post["url"],
                    "inconsistent quote uri found between record.embed.record.uri & embed.record.uri: %s %s"
                    % (post["quoted_uri"], quoted_data),
                )

            quoted_data["record"] = quoted_data["value"]
            del quoted_data["value"]
            if "embeds" in quoted_data and len(quoted_data["embeds"]):
                if len(quoted_data["embeds"]) != 1:
                    raise BlueskyPayloadError(
                        post["url"],
                        "unusual multiple embeds found within a quoted post: %s"
                        % quoted_data["embeds"],
                    )
                quoted_data["embed"] = quoted_data["embeds"][0]
                del quoted_data["embeds"]

            nested = normalize_post(
                quoted_data,
                locale=locale,
                extract_referenced_posts=True,
                collection_source="quote",
            )
            quoted = nested[-1]
            if extract_referenced_posts:
                referenced_posts = merge_nested_posts(
                    referenced_posts, nested, post["url"]
                )

            # Take better quoted url with user_handle
            post["quoted_url"] = quoted["url"]
            post["quoted_user_handle"] = quoted["user_handle"]
            post["quoted_created_at"] = quoted["local_time"]
            post["quoted_timestamp_utc"] = quoted["timestamp_utc"]

            # Remove quoted link from post links if present in text
            if quoted["url"] in links:
                links.remove(quoted["url"])

            # Rewrite post's text to include quote within (or replace the link to the quote if present)
            quote = (
                "« @%s: %s — %s »"
                % (quoted["user_handle"], quoted["text"], quoted["url"])
            ).encode("utf-8")
            url_lower = quoted["url"].encode("utf-8").lower()
            text_lower = text.lower()
            if url_lower in text_lower:
                url_pos = text_lower.find(url_lower)
                text = text[:url_pos] + quote + text[url_pos + len(quoted["url"]) :]
            else:
                text += b" " + quote

    # Process links domains
    post["links"] = sorted(links)
    post["domains"] = [custom_get_normalized_hostname(link) for link in post["links"]]

    # Handle threadgates (replies rules)
    # WARNING: quoted posts do not seem to include threadgates info
    # Issue opened about it here: https://github.com/bluesky-social/atproto/issues/3716
    if "threadgate" in data:
        post["replies_rules"] = []
        if "allow" in data["threadgate"]["record"]:
            for rule in data["threadgate"]["record"]["allow"]:
                rule_string = (
                    "allow_from_" + rule["$type"].split("#")[1].split("Rule")[0]
                )
                if rule_string.endswith("_list") and "list" in rule:
                    for allowed_list in rule["list"]:
                        post["replies_rules"].append(rule_string + ":" + allowed_list)
                else:
                    post["replies_rules"].append(rule_string)
            if not data["threadgate"]["record"]["allow"]:
                post["replies_rules"].append("disallow")
        (
            post["replies_rules_timestamp_utc"],
            post["replies_rules_created_at"],
        ) = get_dates(
            data["threadgate"]["record"]["createdAt"], locale=locale, source="bluesky"
        )
        post["hidden_replies_uris"] = data["threadgate"]["record"].get(
            "hiddenReplies", []
        )

    # Handle postgates (quotes rules)
    #
    # Users can forbid others to quote a post, but payloads do not seem to
    # include it yet although the API spec documents it:
    # https://github.com/bluesky-social/atproto/blob/main/lexicons/app/bsky/feed/postgate.json
    # Issue opened about it here: https://github.com/bluesky-social/atproto/issues/3712
    #
    # if "postgate" in data:
    #     if "embeddingRules" in data["postgate"]["record"] and data["postgate"]["record"]["embeddingRules"]:
    #         post["quotes_rule"] = "disallow"
    #     post["quotes_rules_timestamp_utc"], post["quotes_rules_created_at"] = get_dates(data["postgate"]["record"]["createdAt"], locale=locale, source="bluesky")
    #     post["detached_quotes_uris"] = data["postgate"]["record"].get("detachedEmbeddingUris", [])

    # Handle reposts when data comes from a feed
    if repost_data:
        if not repost_data["$type"].endswith("reasonRepost"):
            raise BlueskyPayloadError(
                post["url"],
                "unusual reason for including a post within a feed: %s" % repost_data,
            )

        post["repost_by_user_did"] = repost_data["by"]["did"]
        post["repost_by_user_handle"] = repost_data["by"]["handle"]
        post["repost_timestamp_utc"], post["repost_created_at"] = get_dates(
            repost_data["indexedAt"], locale=locale, source="bluesky"
        )

    try:
        post["text"] = text.decode("utf-8")
    except UnicodeDecodeError as e:
        raise UnicodeDecodeError(
            e.encoding,
            e.object,
            e.start,
            e.end,
            f"{e.reason} in post {post['url']}.\nText to decode: {text}\nSlice of text to decode: {text[e.start : e.end]}",
        )

    if collection_source is not None:
        post["collected_via"] = [collection_source]
    post["match_query"] = collection_source not in ["thread", "quote"]

    if extract_referenced_posts:
        # Handle thread posts when data comes from a feed
        if reply_data:
            if "parent" in reply_data:
                nested = normalize_post(
                    reply_data["parent"],
                    locale=locale,
                    extract_referenced_posts=True,
                    collection_source="thread",
                )
                referenced_posts = merge_nested_posts(
                    referenced_posts, nested, post["url"]
                )

            if "root" in reply_data and (
                "parent" not in reply_data
                or reply_data["parent"]["cid"] != reply_data["root"]["cid"]
            ):
                nested = normalize_post(
                    reply_data["root"],
                    locale=locale,
                    extract_referenced_posts=True,
                    collection_source="thread",
                )
                referenced_posts = merge_nested_posts(
                    referenced_posts, nested, post["url"]
                )

            if "grandparentAuthor" in reply_data:
                # TODO ? Shall we do anything from that?
                pass

        assert referenced_posts is not None
        return [referenced_posts[did] for did in sorted(referenced_posts.keys())] + [
            post
        ]  # type: ignore

    return post  # type: ignore



# =============================================================================
# Twitwi Normalizers
# =============================================================================
#
# Function used to "normalize" tweets from various sources (v1 API, v2 API,
# public scraped API etc.) in order to produce homogeneous and easily
# analysable data.
#
import re
from copy import deepcopy
from html import unescape

from twitwi.exceptions import TwitterPayloadV2IncompleteIncludesError
from twitwi.utils import (
    get_collection_time,
    get_dates,
    custom_normalize_url,
    validate_payload_v2,
    custom_get_normalized_hostname,
)

CLEAN_RT_PATTERN = re.compile(r"^RT @\w+: ")


def format_rt_text(user, text):
    return "RT @%s: %s" % (user, text)


def format_qt_text(user, text, quoted_text, url):
    clean_url = re.sub(r"(\?s=\d+|/(video|photo)/\d+)+", "", url.lower())
    quote = "« %s: %s — %s »" % (user, quoted_text, clean_url)
    text_lc = text.lower()
    if quote.lower() in text_lc:
        return text
    url_lc = url.lower()
    url_pos = text_lc.find(url_lc)
    if url_pos != -1:
        url_len = len(url)
        return ("%s%s%s" % (text[:url_pos], quote, text[url_pos + url_len :])).strip()
    return "%s %s" % (text, quote)


def format_tweet_url(screen_name, tweet_id):
    return "https://twitter.com/%s/status/%s" % (screen_name, tweet_id)


def extract_media_name_from_url(media_url):
    return media_url.rsplit("/", 1)[-1].split("?tag=", 1)[0]


def extract_items_from_text(text, char):
    splitter = re.compile(r"[^\w%s]+" % char)

    return sorted(
        set(
            r.lstrip(char).lower()
            for r in splitter.split(CLEAN_RT_PATTERN.sub("", text))
            if r.startswith(char)
        )
    )


def extract_hashtags_from_text(text):
    return extract_items_from_text(text, "#")


def extract_mentions_from_text(text):
    return extract_items_from_text(text, "@")


def resolve_entities(tweet, prefix):
    status_key = "%s_status" % prefix
    target = tweet[status_key]

    for ent in ["entities", "extended_entities"]:
        if ent not in target:
            continue
        tweet[ent] = tweet.get(ent, {})
        for field in target[ent]:
            tweet[ent][field] = tweet[ent].get(field, [])
            if field in target[ent]:
                tweet[ent][field] += target[ent][field]


def get_bitrate(x):
    return x.get("bitrate", 0)


def get_bitrate_v2(x):
    return x.get("bit_rate", 0)


def nostr_field(f):
    return f.replace("_str", "")


META_FIELDS = [
    "in_reply_to_status_id_str",
    "in_reply_to_screen_name",
    "in_reply_to_user_id_str",
    "lang",
    "possibly_sensitive",
    "retweet_count",
    "favorite_count",
    "reply_count",
]

META_FIELD_TRANSLATIONS = {
    "in_reply_to_status_id_str": "to_tweetid",
    "in_reply_to_screen_name": "to_username",
    "in_reply_to_user_id_str": "to_userid",
    "favorite_count": "like_count",
}

USER_META_FIELDS = [
    "screen_name",
    "name",
    "friends_count",
    "followers_count",
    "location",
    "verified",
    "description",
    "created_at",
]

PLACE_META_FIELDS = ["country_code", "full_name", "place_type"]


def grab_extra_meta(source, result, locale=None):
    if source.get("coordinates"):
        result["coordinates"] = source["coordinates"]["coordinates"]
        result["lat"] = source["coordinates"]["coordinates"][1]
        result["lng"] = source["coordinates"]["coordinates"][0]
    else:
        # TODO: this is hardly optimal
        result["coordinates"] = None

    for meta in META_FIELDS:
        if meta in source:
            result[META_FIELD_TRANSLATIONS.get(meta, meta)] = source[meta]
        elif nostr_field(meta) in source:
            result[meta] = str(source[nostr_field(meta)])

    # impression_count when scraping
    if "ext_views" in source:
        result["impression_count"] = source["ext_views"].get("count")

    for meta in USER_META_FIELDS:
        key = "user_%s" % meta.replace("_count", "")
        if key in source:
            result[key] = source[key]
        elif "user" in source and meta in source["user"]:
            result[key] = source["user"][meta] if source["user"][meta] != "" else None

    if "user" in source:
        result["user_id"] = source["user"]["id_str"]
        result["user_tweets"] = source["user"]["statuses_count"]
        result["user_likes"] = source["user"]["favourites_count"]
        result["user_lists"] = source["user"]["listed_count"]
        result["user_image"] = source["user"]["profile_image_url_https"]

    if "place" in source and source["place"] is not None:
        for meta in PLACE_META_FIELDS:
            if meta in source["place"]:
                key = "place_%s" % meta.replace("place_", "").replace("full_", "")
                result[key] = source["place"][meta]

        if (
            "bounding_box" in source["place"]
            and source["place"]["bounding_box"] is not None
            and "coordinates" in source["place"]["bounding_box"]
        ):
            result["place_coordinates"] = source["place"]["bounding_box"][
                "coordinates"
            ][0]

    # TODO: nested_get
    try:
        result["user_url"] = source["user"]["entities"]["url"]["urls"][0][
            "expanded_url"
        ]
    except (KeyError, IndexError):
        try:
            result["user_url"] = source["user"]["url"]
        except KeyError:
            pass

    if "user_created_at" in result:
        result["user_timestamp_utc"], result["user_created_at"] = get_dates(
            result["user_created_at"], locale
        )

    if source.get("source"):
        result["source_url"], result["source_name"] = (
            source["source"]
            .replace('<a href="', "")
            .replace("</a>", "")
            .split('" rel="nofollow">')
        )

    return result


def normalize_tweet(
    tweet,
    locale=None,
    extract_referenced_tweets=False,
    collection_source=None,
    pure=True,
):
    """
    Function "normalizing" a tweet as returned by Twitter's API in order to
    cleanup and optimize some fields.

    Note that this function mutates the argument to work.

    Args:
        tweet (dict): Tweet json dict from Twitter API.
        locale (pytz.timezone, optional): Timezone for date conversions.
        extract_referenced_tweets (bool, optional): Whether to return only
            the original tweet or the full list of tweets found in the given
            tweet payload (including quoted and retweeted tweets). Defaults
            to `False`.
        collection_source (str, optional): string explaining how the tweet
            was collected. Defaults to `None`.
        pure (bool, optional): whether to allow the function to mutate its
            original argument. Default to `True`.

    Returns:
        (dict or list): Either a single tweet dict or a list of tweet dicts if
            `extract_referenced_tweets` was set to `True`.

    """

    if pure:
        tweet = deepcopy(tweet)

    results = []

    if "extended_tweet" in tweet:
        for field in tweet["extended_tweet"]:
            tweet[field] = tweet["extended_tweet"][field]

    text = tweet.get("full_text", tweet.get("text", ""))

    rti = None
    rtu = None
    rtuid = None
    rtime = None
    qti = None
    qturl = None
    qtu = None
    qtuid = None
    qtime = None

    if (
        "retweeted_status" in tweet
        and tweet["retweeted_status"]["id_str"] != tweet["id_str"]
    ):
        rti = tweet["retweeted_status"]["id_str"]
        rtu = tweet["retweeted_status"]["user"]["screen_name"]
        rtuid = tweet["retweeted_status"]["user"]["id_str"]

        nested = normalize_tweet(
            tweet["retweeted_status"],
            locale=locale,
            extract_referenced_tweets=True,
            collection_source="retweet",
            pure=False,
        )

        rtweet = nested[-1]

        if extract_referenced_tweets:
            results.extend(nested)

        rtime = rtweet["timestamp_utc"]

        resolve_entities(tweet, "retweeted")

    elif (
        "quoted_status" in tweet and tweet["quoted_status"]["id_str"] != tweet["id_str"]
    ):
        qti = tweet["quoted_status"]["id_str"]
        qtu = tweet["quoted_status"]["user"]["screen_name"]
        qtuid = tweet["quoted_status"]["user"]["id_str"]

        nested = normalize_tweet(
            tweet["quoted_status"],
            locale=locale,
            extract_referenced_tweets=True,
            collection_source="quote",
            pure=False,
        )

        qtweet = nested[-1]

        if extract_referenced_tweets:
            results.extend(nested)

        if "quoted_status_permalink" in tweet:
            qturl = tweet["quoted_status_permalink"]["expanded"]
        else:
            qturl = qtweet["url"]
        qtime = qtweet["timestamp_utc"]

        resolve_entities(tweet, "quoted")

    medids = set()
    media_urls = []
    media_files = []
    media_types = []
    media_alt_texts = []

    links = set()
    hashtags = set()
    mentions = {}

    if "entities" in tweet or "extended_entities" in tweet:
        source_id = rti or qti or tweet["id_str"]

        entities = tweet.get("extended_entities", tweet["entities"]).get("media", [])
        entities += tweet["entities"].get("urls", [])

        for entity in entities:
            if "expanded_url" in entity and "url" in entity and entity["expanded_url"]:
                try:
                    text = text.replace(entity["url"], entity["expanded_url"])
                except KeyError:
                    pass

            if "media_url" in entity or "media_url_https" in entity:
                if "video_info" in entity:
                    med_url = max(entity["video_info"]["variants"], key=get_bitrate)[
                        "url"
                    ]
                else:
                    med_url = entity["media_url_https"]

                med_name = extract_media_name_from_url(med_url)

                if med_name not in medids:
                    medids.add(med_name)
                    media_types.append(entity["type"])
                    media_urls.append(med_url.split("?tag=")[0])
                    media_files.append("%s_%s" % (source_id, med_name))
                    media_alt_texts.append(entity.get("ext_alt_text") or "")

            # NOTE: fun fact, Twitter is starting to break down and we cannot guarantee
            # expanded_url exists anymore. It even crashes the website itself lol:
            # https://x.com/lmerzeau/status/426318495450943488
            elif "expanded_url" in entity:
                normalized = custom_normalize_url(entity["expanded_url"])
                links.add(normalized)

        for hashtag in tweet["entities"].get("hashtags", []):
            hashtags.add(hashtag["text"].lower())

        for mention in tweet["entities"].get("user_mentions", []):
            mentions[mention["screen_name"].lower()] = mention["id_str"]

    if rtu:
        text = format_rt_text(rtu, rtweet["text"])
        if rtweet["quoted_id"]:
            qturl = format_tweet_url(rtweet["quoted_user"], rtweet["quoted_id"])

    elif qtu:
        text = format_qt_text(qtu, text, qtweet["text"], qturl)

    if qturl:
        qturl_lc = custom_normalize_url(qturl).lower()
        for link in list(links):
            if link.lower() == qturl_lc:
                links.remove(link)

    timestamp_utc, local_time = get_dates(tweet["created_at"], locale)
    text = unescape(text)

    if collection_source is None:
        collection_source = tweet.get("collection_source")
    links = sorted(links)
    domains = [
        custom_get_normalized_hostname(
            link, normalize_amp=False, infer_redirection=False
        )
        for link in links
    ]
    normalized_tweet = {
        "id": tweet["id_str"],
        "local_time": local_time,
        "timestamp_utc": timestamp_utc,
        "text": text,
        "url": format_tweet_url(tweet["user"]["screen_name"], tweet["id_str"]),
        "quoted_id": qti,
        "quoted_user": qtu,
        "quoted_user_id": qtuid,
        "quoted_timestamp_utc": qtime,
        "retweeted_id": rti,
        "retweeted_user": rtu,
        "retweeted_user_id": rtuid,
        "retweeted_timestamp_utc": rtime,
        "media_files": media_files,
        "media_types": media_types,
        "media_urls": media_urls,
        "media_alt_texts": media_alt_texts,
        "links": links,
        "links_to_resolve": len(links) > 0,
        "domains": domains,
        "hashtags": sorted(hashtags) if hashtags else extract_hashtags_from_text(text),
        "mentioned_ids": [mentions[m] for m in sorted(mentions.keys())],
        "mentioned_names": sorted(mentions.keys())
        if mentions
        else extract_mentions_from_text(text),
        "collection_time": get_collection_time(),
        "match_query": collection_source != "thread" and collection_source != "quote",
    }

    if collection_source is not None:
        normalized_tweet["collected_via"] = [collection_source]

    grab_extra_meta(tweet, normalized_tweet, locale)

    if rtu and not normalized_tweet["retweet_count"]:
        normalized_tweet["retweet_count"] = rtweet["retweet_count"]

    results.append(normalized_tweet)

    if not extract_referenced_tweets:
        return results[0]

    return results


def resolve_user_entities(user):
    if "entities" in user:
        for k in user["entities"]:
            if "urls" in user["entities"][k]:
                for url in user["entities"][k]["urls"]:
                    if not url.get("expanded_url"):
                        continue
                    if k in user:
                        user[k] = user[k].replace(url["url"], url["expanded_url"])


def normalize_user(user, locale=None, pure=True, v2=False):
    """
    Function "normalizing" a user as returned by Twitter's API in order to
    cleanup and optimize some fields.

    Note that this function mutates the argument to work.

    Args:
        user (dict): Twitter user json dict from Twitter API.
        locale (pytz.timezone, optional): Timezone for date conversions.
        pure (bool, optional): whether to allow the function to mutate its
            original argument. Default to `True`.

    Returns:
        dict: The normalized user.

    """

    if pure:
        user = deepcopy(user)

    resolve_user_entities(user)

    timestamp_utc, local_time = get_dates(
        user["created_at"], locale, source="v2" if v2 else "v1"
    )

    if v2 and "withheld" in user:
        withheld = user["withheld"]
        withheld_in_countries = withheld.get("country_codes", [])
        withheld_scope = withheld.get("withheld_scope", "")
    else:
        withheld_in_countries = []
        withheld_scope = ""

    normalized_user = {
        "id": user["id_str"] if not v2 else user["id"],
        "screen_name": user["screen_name"] if not v2 else user["username"],
        "name": user["name"],
        "description": user["description"] if user["description"] else None,
        "url": user.get("url"),
        "timestamp_utc": timestamp_utc,
        "local_time": local_time,
        "location": user.get("location"),
        "verified": user.get("verified", False),
        "protected": user.get("protected", False),
        "tweets": user["statuses_count"]
        if not v2
        else user["public_metrics"]["tweet_count"],
        "followers": user["followers_count"]
        if not v2
        else user["public_metrics"]["followers_count"],
        "friends": user["friends_count"]
        if not v2
        else user["public_metrics"]["following_count"],
        "likes": user["favourites_count"] if not v2 else None,
        "lists": user["listed_count"]
        if not v2
        else user["public_metrics"]["listed_count"],
        "image": user.get("profile_image_url_https")
        if not v2
        else user.get("profile_image_url"),
        "default_profile": user.get("default_profile", False),
        "default_profile_image": user.get("default_profile_image", False),
        "witheld_in_countries": user.get("witheld_in_countries", [])
        if not v2
        else withheld_in_countries,
        "witheld_scope": user.get("witheld_scope") if not v2 else withheld_scope,
    }

    return normalized_user


def includes_index(payload, key, index_key="id"):
    return {item[index_key]: item for item in payload["includes"].get(key, [])}


def get_best_url(item):
    if "unwound_url" in item:
        return item["unwound_url"]

    if "expanded_url" in item:
        return item["expanded_url"]

    return None


def normalize_tweet_v2(
    tweet,
    *,
    users_by_screen_name,
    places_by_id,
    tweets_by_id,
    users_by_id,
    media_by_key,
    locale=None,
    collection_source=None,
    extract_referenced_tweets=False,
):
    timestamp_utc, local_time = get_dates(
        tweet["created_at"], locale=locale, source="v2"
    )

    try:
        user = users_by_id[tweet["author_id"]]
    except KeyError:
        raise TwitterPayloadV2IncompleteIncludesError("user", tweet["author_id"])

    user_timestamp_utc, user_created_at = get_dates(
        user["created_at"], locale=locale, source="v2"
    )
    user_entities = user.get("entities", {})

    entities = tweet.get("entities", {})
    referenced_tweets = tweet.get("referenced_tweets", [])

    hashtags = set()

    for hashtag in entities.get("hashtags", []):
        hashtags.add(hashtag["tag"])

    mentions = {}

    for mention in entities.get("mentions", []):
        if "id" in mention:
            mentions[mention["username"]] = mention["id"]
        else:
            try:
                mentions[mention["username"]] = users_by_screen_name[
                    mention["username"]
                ]["id"]
            except KeyError:
                raise TwitterPayloadV2IncompleteIncludesError(
                    "user", mention["username"]
                )

    place_info = {}

    if "geo" in tweet:
        geo_data = tweet["geo"]

        if "coordinates" in geo_data:
            point = geo_data["coordinates"]

            if point["type"] == "Point":
                lng, lat = point["coordinates"]
                place_info["lng"] = lng
                place_info["lat"] = lat

        if "place_id" in geo_data:
            place_data = places_by_id.get(geo_data["place_id"], {})

            if "country_code" in place_data:
                place_info["place_country_code"] = place_data["country_code"]

            if "full_name" in place_data:
                place_info["place_name"] = place_data["full_name"]

            if "place_type" in place_data:
                place_info["place_type"] = place_data["place_type"]

            if "geo" in place_data and "bbox" in place_data["geo"]:
                place_info["place_coordinates"] = place_data["geo"]["bbox"]

    # Text
    text = tweet["text"]

    # References
    refs = {t["type"]: t["id"] for t in referenced_tweets}

    # Reply
    reply_info = {}

    if "replied_to" in refs:
        reply = tweets_by_id.get(refs["replied_to"], {})
        if "author_id" in reply:
            try:
                reply_info["to_username"] = users_by_id[reply["author_id"]]["username"]
            except KeyError:
                raise TwitterPayloadV2IncompleteIncludesError(
                    "replied_user", reply["author_id"]
                )
        else:
            reply_info["to_username"] = ""
        reply_info["to_userid"] = reply.get("author_id", "")
        reply_info["to_tweetid"] = reply.get("id", "")

    # Retweet
    retweet_info = {}
    normalized_retweet = None

    if "retweeted" in refs:
        retweet_info["retweeted_id"] = refs["retweeted"]

        if refs["retweeted"] in tweets_by_id:
            retweet = tweets_by_id[refs["retweeted"]]
            normalized_retweet = normalize_tweet_v2(
                retweet,
                users_by_screen_name=users_by_screen_name,
                places_by_id=places_by_id,
                tweets_by_id=tweets_by_id,
                users_by_id=users_by_id,
                media_by_key=media_by_key,
                locale=locale,
                collection_source="retweet",
            )

            retweet_info["retweeted_user"] = normalized_retweet["user_screen_name"]
            retweet_info["retweeted_user_id"] = normalized_retweet["user_id"]
            retweet_info["retweeted_timestamp_utc"] = normalized_retweet[
                "timestamp_utc"
            ]

    # Quoted
    quote_info = {}
    normalized_quote = None

    if "quoted" in refs:
        quote_info["quoted_id"] = refs["quoted"]

        if refs["quoted"] in tweets_by_id:
            quote = tweets_by_id[refs["quoted"]]
            normalized_quote = normalize_tweet_v2(
                quote,
                users_by_screen_name=users_by_screen_name,
                places_by_id=places_by_id,
                tweets_by_id=tweets_by_id,
                users_by_id=users_by_id,
                media_by_key=media_by_key,
                locale=locale,
                collection_source="quote",
            )

            quote_info["quoted_user"] = normalized_quote["user_screen_name"]
            quote_info["quoted_user_id"] = normalized_quote["user_id"]
            quote_info["quoted_timestamp_utc"] = normalized_quote["timestamp_utc"]

    # Replace urls in text
    links = set()

    for url_data in entities.get("urls", []):
        replacement_url = get_best_url(url_data)

        if replacement_url:
            text = text.replace(url_data["url"], replacement_url)

        if not replacement_url:
            replacement_url = url_data["url"]

        links.add(custom_normalize_url(replacement_url))

    if normalized_retweet:
        text = format_rt_text(
            normalized_retweet["user_screen_name"], normalized_retweet["text"]
        )

    if normalized_quote:
        text = format_qt_text(
            normalized_quote["user_screen_name"],
            text,
            normalized_quote["text"],
            normalized_quote["url"],
        )

    # Metrics
    is_retweet = "retweeted" in refs
    public_metrics = tweet["public_metrics"]
    user_public_metrics = user["public_metrics"]

    user_url = user.get("url")

    if "url" in user_entities and "urls" in user_entities["url"]:
        user_url_entity = user_entities["url"]["urls"][0]
        user_url = get_best_url(user_url_entity) or user_url

    # Media
    medias = []

    if "attachments" in tweet and "media_keys" in tweet["attachments"]:
        source_id = refs.get("retweeted_id", tweet["id"])

        for media_key in tweet["attachments"]["media_keys"]:
            if media_key in media_by_key:
                try:
                    media_data = media_by_key[media_key]
                except KeyError:
                    raise TwitterPayloadV2IncompleteIncludesError("media", media_key)

                if "variants" in media_data:
                    media_url = max(media_data["variants"], key=get_bitrate_v2)["url"]
                else:
                    media_url = media_data.get("url", "")
                medias.append(
                    (
                        media_url,
                        "%s_%s" % (source_id, extract_media_name_from_url(media_url)),
                        media_data["type"],
                    )
                )

    if collection_source is None:
        collection_source = tweet.get("collection_source")

    sorted_mentions = sorted(mentions.keys())

    normalized_tweet = {
        "id": tweet["id"],
        "local_time": local_time,
        "timestamp_utc": timestamp_utc,
        "text": unescape(text),
        "url": format_tweet_url(user["username"], tweet["id"]),
        "hashtags": sorted(hashtags),
        "mentioned_names": sorted_mentions,
        "mentioned_ids": [mentions[k] for k in sorted_mentions],
        "collection_time": get_collection_time(),
        "user_id": user["id"],
        "user_screen_name": user["username"],
        "user_name": user["name"],
        "user_image": user["profile_image_url"],
        "user_url": user_url or None,
        "user_location": user.get("location"),
        "user_verified": user["verified"],
        "user_description": user["description"] if user["description"] else None,
        "user_tweets": user_public_metrics.get("tweet_count"),
        "user_followers": user_public_metrics.get("followers_count"),
        "user_friends": user_public_metrics.get("following_count"),
        "user_lists": user_public_metrics.get("listed_count"),
        "user_created_at": user_created_at,
        "user_timestamp_utc": user_timestamp_utc,
        "possibly_sensitive": tweet["possibly_sensitive"],
        "like_count": public_metrics["like_count"] if not is_retweet else 0,
        "retweet_count": public_metrics["retweet_count"] if not is_retweet else 0,
        "quote_count": public_metrics["quote_count"] if not is_retweet else 0,
        "reply_count": public_metrics["reply_count"] if not is_retweet else 0,
        "impression_count": public_metrics.get("impression_count"),
        "lang": tweet["lang"],
        "source_name": tweet.get("source"),
        "links": sorted(links),
        "media_urls": [m[0] for m in medias],
        "media_files": [m[1] for m in medias],
        "media_types": [m[2] for m in medias],
        "match_query": collection_source != "thread" and collection_source != "quote",
        **place_info,
        **reply_info,
        **retweet_info,
        **quote_info,
    }

    if collection_source is not None:
        normalized_tweet["collected_via"] = [collection_source]

    if extract_referenced_tweets:
        normalized_tweets = [normalized_tweet]

        if normalized_retweet is not None:
            normalized_tweets.append(normalized_retweet)

        if normalized_quote is not None:
            normalized_tweets.append(normalized_quote)

        return normalized_tweets

    return normalized_tweet


def normalize_tweets_payload_v2(
    payload, locale=None, extract_referenced_tweets=False, collection_source=None
):
    if not validate_payload_v2(payload):
        raise TypeError("given value is not a Twitter API v2 payload")

    if "data" not in payload:
        return []

    users_by_screen_name = includes_index(payload, "users", index_key="username")
    users_by_id = includes_index(payload, "users")
    places_by_id = includes_index(payload, "places")
    tweets_by_id = includes_index(payload, "tweets")
    media_by_key = includes_index(payload, "media", index_key="media_key")

    output = []
    already_seen = {}

    for item in payload["data"]:
        normalized_tweets = normalize_tweet_v2(
            item,
            users_by_id=users_by_id,
            users_by_screen_name=users_by_screen_name,
            places_by_id=places_by_id,
            tweets_by_id=tweets_by_id,
            media_by_key=media_by_key,
            locale=locale,
            collection_source=collection_source,
            extract_referenced_tweets=True,
        )

        if extract_referenced_tweets:
            for normalized_tweet in normalized_tweets:
                k = int(normalized_tweet["id"])
                earlier_normalized_tweet = already_seen.get(k)

                if earlier_normalized_tweet is not None:
                    if "collected_via" in normalized_tweet:
                        new_collection_source = normalized_tweet["collected_via"][0]

                        if "collected_via" not in earlier_normalized_tweet:
                            earlier_normalized_tweet["collected_via"] = [
                                new_collection_source
                            ]
                        else:
                            if (
                                new_collection_source
                                not in earlier_normalized_tweet["collected_via"]
                            ):
                                earlier_normalized_tweet["collected_via"].append(
                                    new_collection_source
                                )

                    continue

                already_seen[k] = normalized_tweet
                output.append(normalized_tweet)
        else:
            assert len(normalized_tweets)
            output.append(normalized_tweets[0])

    return output



from twitwi.bluesky.normalizers import (
    normalize_profile,
    normalize_partial_profile,
    normalize_post,
)
from twitwi.bluesky.formatters import (
    transform_profile_into_csv_dict,
    format_profile_as_csv_row,
    transform_partial_profile_into_csv_dict,
    format_partial_profile_as_csv_row,
    transform_post_into_csv_dict,
    format_post_as_csv_row,
)

__all__ = [
    "transform_profile_into_csv_dict",
    "format_profile_as_csv_row",
    "transform_partial_profile_into_csv_dict",
    "format_partial_profile_as_csv_row",
    "transform_post_into_csv_dict",
    "format_post_as_csv_row",
    "normalize_profile",
    "normalize_partial_profile",
    "normalize_post",
]



import re

from twitwi.exceptions import BlueskyPayloadError


valid_post_keys = [
    "cid",
    "uri",
    "author",
    "record",
    "replyCount",
    "repostCount",
    "likeCount",
    "quoteCount",
]

valid_record_keys = ["$type", "createdAt", "text"]


valid_author_keys = ["did", "handle", "createdAt"]


def validate_post_payload(data):
    post = data.get("post", data)

    for key in valid_post_keys:
        if key not in post:
            return False, f"key {key} is missing from payload: {post}"

    if not isinstance(post["record"], dict):
        return False, "payload's record is not a dictionary: %s" % post["record"]

    for key in valid_record_keys:
        if key not in post["record"]:
            return False, "key %s is missing from payload's record: %s" % (
                key,
                post["record"],
            )

    if post["record"].get("$type") != "app.bsky.feed.post":
        return False, "payload's record $type is not a post: %s" % post["record"].get(
            "$type"
        )

    if not isinstance(post["author"], dict):
        return False, "payload's author is not a dictionary: %s" % post["author"]

    for key in valid_author_keys:
        if key not in post["author"]:
            return False, "key %s is missing from payload's author: %s" % (
                key,
                post["author"],
            )

    return True, None


re_embed_types = re.compile(
    r"\.(record|recordWithMedia|images|video|external)(?:#.*)?$"
)


def valid_embed_type(embed_type):
    return re_embed_types.search(embed_type)


def format_profile_url(user_handle_or_did):
    return f"https://bsky.app/profile/{user_handle_or_did}"


def format_post_url(user_handle_or_did, post_did, post_splitter="/post/"):
    return f"https://bsky.app/profile/{user_handle_or_did}{post_splitter}{post_did}"


def parse_post_url(url, source):
    """Returns a tuple of (author_handle/did, post_did) from an https://bsky.app post URL"""

    known_splits = ["/post/", "/lists/", "/feed/"]

    if url.startswith("https://bsky.app/profile/"):
        for split in known_splits:
            if split in url[25:]:
                return url[25:].split(split)

    raise BlueskyPayloadError(source, f"{url} is not a usual Bluesky post url")


def parse_post_uri(uri, source=None):
    """Returns a tuple of (author_did, post_did) from an at:// post URI"""

    known_splits = [
        "/app.bsky.feed.post/",
        "/app.bsky.graph.starterpack/",
        "/app.bsky.feed.generator/",
        "/app.bsky.graph.list/",
    ]

    if uri.startswith("at://"):
        for split in known_splits:
            if split in uri:
                return uri[5:].split(split)

    raise BlueskyPayloadError(source or uri, f"{uri} is not a usual Bluesky post uri")


def format_starterpack_url(user_handle_or_did, record_did):
    return f"https://bsky.app/starter-pack/{user_handle_or_did}/{record_did}"


def format_media_url(user_did, media_cid, mime_type, source):
    media_type = mime_type.split("/")[1]
    if mime_type.startswith("image"):
        media_url = f"https://cdn.bsky.app/img/feed_fullsize/plain/{user_did}/{media_cid}@{media_type}"
        media_thumb = f"https://cdn.bsky.app/img/feed_thumbnail/plain/{user_did}/{media_cid}@{media_type}"
    elif mime_type.startswith("video"):
        media_url = f"https://video.bsky.app/watch/{user_did}/{media_cid}/playlist.m3u8"
        media_thumb = (
            f"https://video.bsky.app/watch/{user_did}/{media_cid}/thumbnail.jpg"
        )
    elif mime_type in ["application/octet-stream", "text/plain"]:
        media_url = (
            f"https://cdn.bsky.app/img/feed_fullsize/plain/{user_did}/{media_cid}@jpeg"
        )
        media_thumb = (
            f"https://cdn.bsky.app/img/feed_thumbnail/plain/{user_did}/{media_cid}@jpeg"
        )
    else:
        raise BlueskyPayloadError(source, f"{mime_type} is an unusual media mimeType")
    return media_url, media_thumb



# =============================================================================
# Twitwi Utilities
# =============================================================================
#
# Miscellaneous utility functions.
#
from typing import Tuple

from pytz import timezone
from dateutil.parser import parse as parse_date
from ural import normalize_url, get_normalized_hostname
from functools import partial
from datetime import datetime

from twitwi.constants import (
    SOURCE_DATETIME_FORMAT,
    SOURCE_DATETIME_FORMAT_V2,
    FORMATTED_TWEET_DATETIME_FORMAT,
    FORMATTED_FULL_DATETIME_FORMAT,
    CANONICAL_URL_KWARGS,
    CANONICAL_HOSTNAME_KWARGS,
    PRE_SNOWFLAKE_LAST_TWEET_ID,
    OFFSET_TIMESTAMP,
)

UTC_TIMEZONE = timezone("UTC")

custom_normalize_url = partial(normalize_url, **CANONICAL_URL_KWARGS)


def safe_normalize_url(url):
    # We avoid normalizing bluesky urls containing specific uri parts because
    # Bluesky servers don't handle quoting correctly...
    # See https://github.com/medialab/twitwi/issues/72
    if "/did:plc:" in url:
        return url

    try:
        return custom_normalize_url(url)
    except Exception:
        # In case of error, return the original URL. Possibly not a valid URL, e.g. url containing double slashes
        return url


custom_get_normalized_hostname = partial(
    get_normalized_hostname, **CANONICAL_HOSTNAME_KWARGS
)


def get_collection_time():
    return datetime.now().strftime(FORMATTED_FULL_DATETIME_FORMAT)


def get_dates(
    date_str: str, locale=None, source: str = "v1", millisecond_timestamp: bool = False
) -> Tuple[int, str]:
    if source not in ["v1", "v2", "bluesky"]:
        raise Exception("source should be one of v1, v2 or bluesky")

    if locale is None:
        locale = UTC_TIMEZONE

    # Let's pray we never see a negative year...
    year_zero = date_str.startswith("0000")

    try:
        parsed_datetime = datetime.strptime(
            date_str,
            SOURCE_DATETIME_FORMAT if source == "v1" else SOURCE_DATETIME_FORMAT_V2,
        )
    except ValueError as e:
        if source != "bluesky":
            raise e
        # Yes, it seems that some people were active in year 0...
        # see by yourself: https://bsky.app/profile/koro.icu/post/3kbpuogc6fz2o
        if year_zero:
            date_str_fixed = "0001" + date_str[4:]
            parsed_datetime = parse_date(date_str_fixed)
        else:
            parsed_datetime = parse_date(date_str)

    utc_datetime = parsed_datetime
    if not parsed_datetime.tzinfo:
        utc_datetime = UTC_TIMEZONE.localize(parsed_datetime)
    locale_datetime = utc_datetime.astimezone(locale)

    timestamp = int(utc_datetime.timestamp())

    if year_zero:
        # Subtract one year (year 0001 is not a leap year) in seconds
        timestamp -= 31536000

    if millisecond_timestamp:
        timestamp *= 1000
        timestamp += utc_datetime.microsecond / 1000

    formatted_date_str = datetime.strftime(
        locale_datetime,
        FORMATTED_FULL_DATETIME_FORMAT
        if source == "bluesky"
        else FORMATTED_TWEET_DATETIME_FORMAT,
    )

    return (
        int(timestamp),
        formatted_date_str if not year_zero else "0" + formatted_date_str[1:],
    )


def validate_payload_v2(payload):
    if not isinstance(payload, dict):
        return False

    if "data" not in payload:
        if (
            "meta" in payload
            and "result_count" in payload["meta"]
            and payload["meta"]["result_count"] == 0
        ):
            return True
        else:
            return False

    if not isinstance(payload["data"], list):
        return False

    # NOTE: not sure it cannot be absent altogether
    if "includes" not in payload or not isinstance(payload["includes"], dict):
        return False

    return True


def get_timestamp_from_id(tweet_id):
    tweet_id = int(tweet_id)

    if tweet_id > PRE_SNOWFLAKE_LAST_TWEET_ID:
        timestamp = (tweet_id >> 22) + OFFSET_TIMESTAMP
        return int(timestamp / 1000)

    return None


def get_dates_from_id(tweet_id, locale=None):
    if locale is None:
        locale = UTC_TIMEZONE

    timestamp = get_timestamp_from_id(tweet_id)
    assert timestamp is not None

    locale_datetime = datetime.fromtimestamp(timestamp, locale)

    return (
        timestamp,
        datetime.strftime(locale_datetime, FORMATTED_TWEET_DATETIME_FORMAT),
    )



import csv
from io import StringIO
from twitwi.bluesky import (
    format_profile_as_csv_row,
    format_partial_profile_as_csv_row,
    format_post_as_csv_row,
    transform_profile_into_csv_dict,
    transform_partial_profile_into_csv_dict,
    transform_post_into_csv_dict,
)
from twitwi.bluesky.constants import PROFILE_FIELDS, PARTIAL_PROFILE_FIELDS, POST_FIELDS
from test.utils import get_json_resource, open_resource


# Set to True to regenerate test results
OVERWRITE_TESTS = False


class TestFormatters:
    def test_format_profile_as_csv_row(self):
        normalized_profiles = get_json_resource("bluesky-normalized-profiles.json")

        buffer = StringIO(newline=None)
        writer = csv.writer(buffer, quoting=csv.QUOTE_MINIMAL)
        writer.writerow(PROFILE_FIELDS)

        for profile in normalized_profiles:
            writer.writerow(format_profile_as_csv_row(profile))

        if OVERWRITE_TESTS:
            written = buffer.getvalue()

            with open("test/resources/bluesky-profiles-export.csv", "w") as f:
                f.write(written)

        with open_resource("bluesky-profiles-export.csv") as f:
            buffer.seek(0)
            assert list(csv.reader(buffer)) == list(csv.reader(f))

    def test_transform_profile_into_csv_dict(self):
        normalized_profiles = get_json_resource("bluesky-normalized-profiles.json")

        buffer = StringIO(newline=None)
        writer = csv.DictWriter(
            buffer,
            fieldnames=PROFILE_FIELDS,
            extrasaction="ignore",
            restval="",
            quoting=csv.QUOTE_MINIMAL,
        )
        writer.writeheader()

        for profile in normalized_profiles:
            transform_profile_into_csv_dict(profile)
            writer.writerow(profile)

        with open_resource("bluesky-profiles-export.csv") as f:
            buffer.seek(0)
            assert list(csv.DictReader(buffer)) == list(csv.DictReader(f))

    def test_format_partial_profile_as_csv_row(self):
        normalized_partial_profiles = get_json_resource(
            "bluesky-normalized-partial-profiles.json"
        )

        buffer = StringIO(newline=None)
        writer = csv.writer(buffer, quoting=csv.QUOTE_MINIMAL)
        writer.writerow(PARTIAL_PROFILE_FIELDS)

        for profile in normalized_partial_profiles:
            writer.writerow(format_partial_profile_as_csv_row(profile))

        if OVERWRITE_TESTS:
            written = buffer.getvalue()

            with open("test/resources/bluesky-partial-profiles-export.csv", "w") as f:
                f.write(written)

        with open_resource("bluesky-partial-profiles-export.csv") as f:
            buffer.seek(0)
            assert list(csv.reader(buffer)) == list(csv.reader(f))

    def test_transform_partial_profile_into_csv_dict(self):
        normalized_partial_profiles = get_json_resource(
            "bluesky-normalized-partial-profiles.json"
        )

        buffer = StringIO(newline=None)
        writer = csv.DictWriter(
            buffer,
            fieldnames=PARTIAL_PROFILE_FIELDS,
            extrasaction="ignore",
            restval="",
            quoting=csv.QUOTE_MINIMAL,
        )
        writer.writeheader()

        for profile in normalized_partial_profiles:
            transform_partial_profile_into_csv_dict(profile)
            writer.writerow(profile)

        with open_resource("bluesky-partial-profiles-export.csv") as f:
            buffer.seek(0)
            assert list(csv.DictReader(buffer)) == list(csv.DictReader(f))

    def test_format_post_as_csv_row(self):
        normalized_posts = get_json_resource("bluesky-normalized-posts.json")

        buffer = StringIO(newline=None)
        writer = csv.writer(buffer, quoting=csv.QUOTE_MINIMAL)
        writer.writerow(POST_FIELDS)

        for source in normalized_posts:
            for post in source:
                writer.writerow(format_post_as_csv_row(post))

        if OVERWRITE_TESTS:
            written = buffer.getvalue()

            with open("test/resources/bluesky-posts-export.csv", "w") as f:
                f.write(written)

        with open_resource("bluesky-posts-export.csv") as f:
            buffer.seek(0)

            assert list(csv.reader(buffer)) == list(csv.reader(f))

    def test_transform_post_into_csv_dict(self):
        normalized_posts = get_json_resource("bluesky-normalized-posts.json")

        buffer = StringIO(newline=None)
        writer = csv.DictWriter(
            buffer,
            fieldnames=POST_FIELDS,
            extrasaction="ignore",
            restval="",
            quoting=csv.QUOTE_MINIMAL,
        )
        writer.writeheader()

        for source in normalized_posts:
            for post in source:
                transform_post_into_csv_dict(post)
                writer.writerow(post)

        with open_resource("bluesky-posts-export.csv") as f:
            buffer.seek(0)
            assert list(csv.DictReader(buffer)) == list(csv.DictReader(f))



import re

QUOTED_REDACT_RE = re.compile(r"«\s+[^»]+:\s+([^»]+)\s+»")


def redact_quoted_text(text: str) -> str:
    return QUOTED_REDACT_RE.sub(r"« \1 »", text)


def redact_rt_text(text: str) -> str:
    return "RT: " + text.split(": ", 1)[1]


FIELDS_TO_DELETE = [
    # The tweet's url leaks the user
    "url",
    # User's place
    "lat",
    "lng",
    "place_coordinates",
    "place_country_code",
    "place_name",
    "place_type",
    "user_location",
    # User info
    "user_created_at",
    "user_description",
    "user_id",
    "user_image",
    "user_name",
    "user_screen_name",
    "user_timestamp_utc",
    "user_url",
    "user_verified",
    # Retweeted user info
    "retweeted_timestamp_utc",
    "retweeted_user",
    "retweeted_user_id",
    # Replied user info
    "to_userid",
    "to_username",
    # Quoted user info
    "quoted_user",
    "quoted_user_id",
    "quoted_timestamp_utc",
]


# NOTE: currently we still keep the id, but we should drop it
# to really call this an anonymized tweet.
# NOTE: we do not redact mentions either.
# NOTE: we also don't redact replies.
def anonymize_normalized_tweet(normalized_tweet) -> None:
    # Text mangling
    text = normalized_tweet["text"]

    if normalized_tweet.get("retweeted_id", None) is not None:
        normalized_tweet["text"] = redact_rt_text(text)

    elif normalized_tweet.get("quoted_id", None) is not None:
        normalized_tweet["text"] = redact_quoted_text(text)

    for field in FIELDS_TO_DELETE:
        if field in normalized_tweet:
            del normalized_tweet[field]



from typing import TypedDict, List, Optional


# To know more about Bluesky IDs:
# - https://docs.bsky.app/docs/advanced-guides/resolving-identities
# - https://bluesky.idunno.dev/docs/commonTerms.html

class BlueskyProfile(TypedDict):
    did: str                            # persistent long-term identifier of the account
    url: str                            # URL of the profile accessible on the web
    handle: str                         # updatable human-readable username of the account (usually like username.bsky.social or username.com)
    display_name: Optional[str]         # updatable human-readable name of the account
    description: Optional[str]          # profile short description written by the user
    posts: int                          # total number of posts submitted by the user (at collection time)
    followers: int                      # total number of followers of the user (at collection time)
    follows: int                        # total number of other users followed by the user (at collection time)
    lists: int                          # total number of lists created by the user (at collection time)
    feedgens: int                       # total number of custom feeds created by the user (at collection time)
    starter_packs: int                  # total number of starter packs created by the user (at collection time)
    avatar: Optional[str]               # URL to the image serving as avatar to the user
    banner: Optional[str]               # URL to the image serving as profile banner to the user
    pinned_post_uri: Optional[str]      # ATProto's internal URI to the post potentially pinned by the user to appear at the top of his posts on his profile
    created_at: str                     # datetime (potentially timezoned) of when the user created the account
    timestamp_utc: int                  # Unix UTC timestamp of when the user created the account
    collection_time: Optional[str]      # datetime (potentially timezoned) of when the data was normalized

class BlueskyPartialProfile(TypedDict): # A partial version of the profile found in follower/follow profile payloads
    did: str                            # persistent long-term identifier of the account
    url: str                            # URL of the profile accessible on the web
    handle: str                         # updatable human-readable username of the account (usually like username.bsky.social or username.com)
    display_name: Optional[str]         # updatable human-readable name of the account
    description: Optional[str]          # profile short description written by the user
    lists: Optional[int]                # total number of lists created by the user (at collection time)
    feedgens: Optional[int]             # total number of custom feeds created by the user (at collection time)
    starter_packs: Optional[int]        # total number of starter packs created by the user (at collection time)
    avatar: Optional[str]               # URL to the image serving as avatar to the user
    created_at: str                     # datetime (potentially timezoned) of when the user created the account
    timestamp_utc: int                  # Unix UTC timestamp of when the user created the account
    collection_time: Optional[str]      # datetime (potentially timezoned) of when the data was normalized



class BlueskyPost(TypedDict):
    # Identifying fields
    cid: str                            # internal content identifier of the post
    did: str                            # persistent long-term identifier of the post
    uri: str                            # ATProto's internal URI to the post
    url: str                            # URL of the post accessible on the web

    # Datetime fields
    timestamp_utc: int                  # Unix UTC timestamp of when the post was submitted
    local_time: str                     # datetime (potentially timezoned) of when the post was submitted
    indexed_at_utc: str                 # datetime (NOT timezoned, for reuse of the Bluesky API) of when the post was indexed by the Bluesky service

    # Author identifying fields
    user_did: str                       # persistent long-term identifier of the account who authored the post
    user_handle: str                    # updatable human-readable username of the account who authored the post

    # Content fields
    text: str                           # reprocessed complete text of the post, including full links, full text of the quoted post (recursively up to 3 posts) and links to images or videos included within
    original_text: str                  # original text of the post as returned by the Bluesky API

    # Metrics fields
    repost_count: int                   # total number of reposts of the post (at collection time)
    like_count: int                     # total number of likes received by the post (at collection time)
    reply_count: int                    # total number of replies received by the post (at collection time)
    quote_count: int                    # total number of posts the post was quoted into (at collection time)
    bookmark_count: Optional[int]       # total number of bookmarks received by the post (at collection time)

    # Extra field
    bridgy_original_url: Optional[str]  # source of the original post when it was posted from another platform such as Mastodon via the Bridgy connection tool

    # Author metadata fields
    user_url: str                       # URL of the profile accessible on the web of the account who authored the post
    user_diplay_name: str               # updatable human-readable name of the account who authored the post
    # user_description: str             # not available from posts payloads
    # user_posts: int                   # not available from posts payloads
    # user_followers: int               # not available from posts payloads
    # user_follows: int                 # not available from posts payloads
    # user_lists: int                   # not available from posts payloads
    user_langs: List[str]               # languages in which the author of the posts usually writes posts (declarative)
    user_avatar: Optional[str]          # URL to the image serving as avatar to the user who authored the post
    user_created_at: str                # datetime (potentially timezoned) of when the user who authored the post created the account
    user_timestamp_utc: int             # Unix UTC timestamp of when the user who authored the post created the account

    # Parent post identifying fields
    # (if the post comes in a conversation as an answer to another post)
    to_post_cid: Optional[str]          # internal content identifier of the parent post
    to_post_did: Optional[str]          # persistent long-term identifier of the parent post
    to_post_uri: Optional[str]          # ATProto's internal URI of the parent post
    to_post_url: Optional[str]          # URL of the parent post on the web
    to_user_did: Optional[str]          # persistent long-term identifier of the account who authored the parent post
    # to_user_handle: Optional[str]     # not available from posts payloads, cf https://github.com/bluesky-social/atproto/issues/3722

    # Root conversation post identifying fields
    # (if the post comes in a conversation, this post is the first one that initiated the thread)
    to_root_post_cid: Optional[str]         # internal content identifier of the root conversation post
    to_root_post_did: Optional[str]         # persistent long-term identifier of the root conversation post
    to_root_post_uri: Optional[str]         # ATProto's internal URI of the root conversation post
    to_root_post_url: Optional[str]         # URL of the root conversation post on the web
    to_root_user_did: Optional[str]         # persistent long-term identifier of the account who authored the root conversation post
    # to_root_user_handle: Optional[str]    # not available from posts payloads, cf https://github.com/bluesky-social/atproto/issues/3722

    # Repost metadata fields
    # Contrary to Twitter where a retweet is a new tweet with its own ID, reposting on Bluesky only adds a flag to the post saying it was reposted by a specific user at a specific time. These are available for instance when collecting a user's feed.
    repost_by_user_did: Optional[str]       # persistent long-term identifier of the account who reposted the post
    repost_by_user_handle: Optional[str]    # updatable human-readable username of the account who reposted the post
    repost_created_at: Optional[int]        # datetime (potentially timezoned) of when the repost was done
    repost_timestamp_utc: Optional[int]     # Unix UTC timestamp of when the repost was done

    # Quoted post metadata fields
    # (when a post embeds another one)
    quoted_cid: Optional[str]           # internal content identifier of the quoted post
    quoted_did: Optional[str]           # persistent long-term identifier of the quoted post
    quoted_uri: Optional[str]           # ATProto's internal URI to the quoted post
    quoted_url: Optional[str]           # URL of the quoted post accessible on the web
    quoted_user_did: Optional[str]      # persistent long-term identifier of the account who authored the quoted post
    quoted_user_handle: Optional[str]   # updatable human-readable username of the account who authored the quoted post
    quoted_created_at: Optional[int]    # datetime (potentially timezoned) of when the quoted post was submitted
    quoted_timestamp_utc: Optional[int] # Unix UTC timestamp of when the quoted post was submitted
    quoted_status: Optional[str]        # empty or "detached" when the author of the quoted post intentionnally required the quoting post not to appear in the list of this post's quotes

    # Embedded elements metadata fields
    links: List[str]                    # list of URLs of all links shared within the post (including potentially the embedded card detailed below, but not the link to a potential quoted post)
    domains: List[str]                  # list of domains of the links shared within the post (here a domain refers to a full hostname, including subdomains, for instance bluesky.com or medialab.sciencespo.fr)
    card_link: Optional[str]            # URL of the link displayed as a card within the post if any
    card_title: Optional[str]           # title of the webpage corresponding to the linkg diplayed as a card within the post if any
    card_description: Optional[str]     # description of the webpage corresponding to the linkg diplayed as a card within the post if any
    card_thumbnail: Optional[str]       # image displayed as an illustration of the webpage corresponding to the link diplayed as a card within the post if any
    media_urls: List[str]               # list of URLs to all media (images, videos, gifs) embedded in the post
    media_thumbnails: List[str]         # list of URLs to small thumbnail version of all media (images, videos, gifs) embedded in the post
    media_types: List[str]              # MIME types (such as image/jpeg, image/gif, video/mp4, etc.) of all media (images, videos, gifs) embedded in the post
    media_alt_texts: List[str]          # description texts of all media (images, videos, gifs) embedded in the post
    mentioned_user_dids: List[str]      # list of all persistent long-term identifiers of the accounts adressed within the post (does not include users to which the post replied)
    mentioned_user_handles: List[str]   # list of all updatable human-readable usernames of the accounts adressed within the post (does not include users to which the post replied)
    hashtags: List[str]                 # list of all unique lowercased hashtags found within the post's text

    # Conversation rules fields
    replies_rules: Optional[List[str]]          # list of specific conversation rules set by the author for the current post (can be one or a combination of: disallow, allow_from_follower, allow_from_following, allow_from_mention, or allow_from_list: followed by a list of user DIDs)
    replies_rules_created_at: Optional[str]     # datetime (potentially timezoned) of when the user set the replies_rules
    replies_rules_timestamp_utc: Optional[int]  # Unix UTC timestamp of when the user set the replies_rules
    hidden_replies_uris: Optional[List[str]]    # list of ATProto's internal URIs to posts who replied to the post, but where intentionnally marked as hidden by the current post's author
    # quotes_rule: Optional[str]                # not available from posts payloads, cf https://github.com/bluesky-social/atproto/issues/3712
    # quotes_rules_created_at: Optional[str]    # not available from posts payloads, cf https://github.com/bluesky-social/atproto/issues/3712
    # quotes_rules_timestamp_utc: Optional[int] # not available from posts payloads, cf https://github.com/bluesky-social/atproto/issues/3712
    # detached_quotes: Optional[List[str]]      # not available from posts payloads, cf https://github.com/bluesky-social/atproto/issues/3712

    # Extra fields linked to the data collection and processing
    collection_time: Optional[str]      # datetime (potentially timezoned) of when the data was normalized
    collected_via: Optional[List[str]]  # extra field added by the normalization process to express how the data collection was ran, will be "quote" or "thread" when a post was grabbed as a referenced post within the originally collected post using the "extract_referenced_posts" option of "normalize_post"
    match_query: Optional[bool]         # extra field added by the normalization process to express whether the post was an intentionnally collected one or only came as a referenced post within the originally collected post using the "extract_referenced_posts" option of "normalize_post"



from setuptools import setup, find_packages

with open("./README.md", "r") as f:
    long_description = f.read()

setup(
    name="twitwi",
    version="0.23.0",
    description="A collection of Twitter-related helper functions for python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/medialab/twitwi",
    license="MIT",
    author="Béatrice Mazoyer, Guillaume Plique, Benjamin Ooghe-Tabanou",
    author_email="guillaume.plique@sciencespo.fr",
    keywords="twitter",
    python_requires=">=3.8",
    packages=find_packages(exclude=["scripts", "test"]),
    package_data={"docs": ["README.md"]},
    install_requires=["pytz>=2019.3", "ural>=0.31.1", "python-dateutil>=2.9.0"],
    zip_safe=True,
)



from twitwi.formatters import make_transform_into_csv_dict, make_format_as_csv_row
from twitwi.bluesky.constants import (
    PROFILE_FIELDS,
    PARTIAL_PROFILE_FIELDS,
    POST_FIELDS,
    POST_PLURAL_FIELDS,
    POST_BOOLEAN_FIELDS,
)


transform_post_into_csv_dict = make_transform_into_csv_dict(
    POST_PLURAL_FIELDS, POST_BOOLEAN_FIELDS
)

format_post_as_csv_row = make_format_as_csv_row(
    POST_FIELDS, POST_PLURAL_FIELDS, POST_BOOLEAN_FIELDS
)


transform_profile_into_csv_dict = make_transform_into_csv_dict([], [])

format_profile_as_csv_row = make_format_as_csv_row(PROFILE_FIELDS, [], [])

transform_partial_profile_into_csv_dict = make_transform_into_csv_dict([], [])

format_partial_profile_as_csv_row = make_format_as_csv_row(
    PARTIAL_PROFILE_FIELDS, [], []
)


__all__ = [
    "transform_post_into_csv_dict",
    "format_post_as_csv_row",
    "transform_profile_into_csv_dict",
    "format_profile_as_csv_row",
    "transform_partial_profile_into_csv_dict",
    "format_partial_profile_as_csv_row",
]
